/*	$FreeBSD$	*/

/*-
 * Copyright (c) 2006
 *	Damien Bergamini <damien.bergamini@free.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#if ( __FreeBSD_version > 700000 )
#include <net80211/ieee80211_amrr.h>
#else
#include <dev/wpi/ieee80211_amrr.h>
#endif

/* DMA mapping */
struct wpi_mapping {
	int nsegs;
	bus_dma_segment_t segs[WPI_MAX_SCATTER];
};

struct wpi_rx_radiotap_header {
	struct ieee80211_radiotap_header wr_ihdr;
	uint64_t	wr_tsft;
	uint8_t		wr_flags;
	uint8_t		wr_rate;
	uint16_t	wr_chan_freq;
	uint16_t	wr_chan_flags;
	int8_t		wr_dbm_antsignal;
	int8_t		wr_dbm_antnoise;
	uint8_t		wr_antenna;
};

#define WPI_RX_RADIOTAP_PRESENT						\
	((1 << IEEE80211_RADIOTAP_TSFT) |				\
	 (1 << IEEE80211_RADIOTAP_FLAGS) |				\
	 (1 << IEEE80211_RADIOTAP_RATE) |				\
	 (1 << IEEE80211_RADIOTAP_CHANNEL) |				\
	 (1 << IEEE80211_RADIOTAP_DBM_ANTSIGNAL) |			\
	 (1 << IEEE80211_RADIOTAP_DBM_ANTNOISE) |			\
	 (1 << IEEE80211_RADIOTAP_ANTENNA))

struct wpi_tx_radiotap_header {
	struct ieee80211_radiotap_header wt_ihdr;
	uint8_t		wt_flags;
	uint8_t		wt_rate;
	uint16_t	wt_chan_freq;
	uint16_t	wt_chan_flags;
	uint8_t		wt_hwqueue;
};

#define WPI_TX_RADIOTAP_PRESENT						\
	((1 << IEEE80211_RADIOTAP_FLAGS) |				\
	 (1 << IEEE80211_RADIOTAP_RATE) |				\
	 (1 << IEEE80211_RADIOTAP_CHANNEL))

struct wpi_dma_info {
	bus_dma_tag_t		tag;
	bus_dmamap_t            map;
	bus_addr_t		paddr;
	caddr_t			vaddr;
	bus_size_t		size;
};

struct wpi_tx_data {
	bus_dmamap_t		map;
	struct mbuf		*m;
	struct ieee80211_node	*ni;
};

struct wpi_tx_ring {
	struct wpi_dma_info	desc_dma;
	struct wpi_dma_info	cmd_dma;
	struct wpi_tx_desc	*desc;
	struct wpi_tx_cmd	*cmd;
	struct wpi_tx_data	*data;
	bus_dma_tag_t		data_dmat;
	int			qid;
	int			count;
	int			queued;
	int			cur;
};

#define WPI_RBUF_COUNT ( WPI_RX_RING_COUNT + 16 )

struct wpi_softc;

struct wpi_rbuf {
	struct wpi_softc	*sc;
	bus_addr_t		paddr;
	caddr_t			vaddr;
	SLIST_ENTRY(wpi_rbuf)	next;
};

struct wpi_rx_data {
	struct mbuf	*m;
};

struct wpi_rx_ring {
	struct wpi_dma_info	desc_dma;
	struct wpi_dma_info	buf_dma;
	bus_dma_tag_t		data_dmat;
	uint32_t		*desc;
	struct wpi_rx_data	data[WPI_RX_RING_COUNT];
	struct wpi_rbuf		rbuf[WPI_RBUF_COUNT];
	SLIST_HEAD(, wpi_rbuf)	freelist;
	int			cur;
};

struct wpi_amrr {
	struct	ieee80211_node ni;	/* must be the first */
	int	txcnt;
	int	retrycnt;
	int	success;
	int	success_threshold;
	int	recovery;
};

struct wpi_node {
        struct  ieee80211_node ni;      /* must be the first */
        struct  ieee80211_amrr_node     amn;
};

struct wpi_softc {
	device_t		sc_dev;
	struct ifnet		*sc_ifp;
	struct ieee80211com	sc_ic;
	int			(*sc_newstate)(struct ieee80211com *,
				    enum ieee80211_state, int);

	struct mtx		sc_mtx;

	struct ieee80211_amrr	amrr;

	uint32_t		sc_flags;
#define WPI_FLAG_INIT_LOCKED	(1 << 0)

	/* shared area */
	struct wpi_dma_info	shared_dma;
	struct wpi_shared	*shared;

	struct wpi_tx_ring	txq[4];
	struct wpi_tx_ring	cmdq;
	struct wpi_tx_ring	svcq;
	struct wpi_rx_ring	rxq;

	struct callout		amrr_ch;

	struct resource		*irq;
	struct resource		*mem;
	bus_space_tag_t		sc_st;
	bus_space_handle_t	sc_sh;
	void 			*sc_ih;
	int			mem_rid;
	int			irq_rid;

	struct wpi_config	config;
	uint16_t		pwr1[14];
	uint16_t		pwr2[14];

	int			sc_tx_timer;

	struct bpf_if		*sc_drvbpf;

	struct wpi_rx_radiotap_header sc_rxtap;
	int			sc_rxtap_len;
	struct wpi_tx_radiotap_header sc_txtap;
	int			sc_txtap_len;
       uint8_t                *sc_boot;
       int                    sc_bootsz;

       bus_dma_tag_t          sc_text_dma_tag;
       bus_dmamap_t           sc_text_dma_map;
       uint8_t                *sc_text;
       int                    sc_textsz;
       struct wpi_tx_desc     sc_text_desc;

       bus_dma_tag_t          sc_data_dma_tag;
       bus_dmamap_t           sc_data_dma_map;
       uint8_t                *sc_data;
       int                    sc_datasz;
       struct wpi_tx_desc     sc_data_desc;
};

#define WPI_LOCK(sc)   mtx_lock(&(sc)->sc_mtx)
#define WPI_UNLOCK(sc) mtx_unlock(&(sc)->sc_mtx)
